import os
import shutil
import logging

from zoom_io import run_command, posixify, relative_path

class ZoomResource(object):
    '''
    Base class for resources
    '''
    def __init__(self, root, path, url):
        '''
        Initializes zoom resource from serialized state.

        args:
            root: root of the ZoomProject
            path: path to a resource (relative path)
            url: url of resource to track
        '''
        self.rtype = None
        self.root = root
        self.path = posixify(path)
        self.url = url
        self.abspath = os.path.normpath(os.path.join(self.root, path))
        # identifier will be used to check for uniqueness amongst resources
        # up to the child class to define
        self.identifier = None
        self._log = logging.getLogger('zoom')

    def __str__(self):
        '''
        String representation of this object.

        returns:
            str: string representation of this object
        '''
        return "{0}::{1}".format(self.rtype, self.identifier)

    def is_synced(self):
        '''
        Check if this resource is already synced
        '''
        raise NotImplementedError()

    def sync(self):
        '''
        Sync this resource to the correct state.
        '''
        raise NotImplementedError()

    def status(self):
        '''
        String representation of this module's status
        '''
        raise NotImplementedError()

    def delete(self):
        '''
        Remove this resource from the filesystem.
        '''
        shutil.rmtree(self.abspath, ignore_errors=True)

    def serialize(self):
        '''
        Serialize state of this resource.
        '''
        raise NotImplementedError()


class ZoomDependency(ZoomResource):
    '''
    Base class for dependencies (files)
    '''
    def __init__(self, root, path, url):
        '''
        Initializes zoom dependency from serialized state.

        args:
            root: root of the ZoomProject
            path: path to the dependency (relative path)
            url: url of the dependency
        '''
        # correct path
        if not path:
            path = relative_path(root)
        super(ZoomDependency, self).__init__(root, path, url)


class ZoomModule(ZoomResource):
    '''
    Base class for modules (scm repos)
    '''
    def __init__(self, root, path, url):
        '''
        Initializes zoom module from serialized state.

        args:
            root: root of the ZoomProject
            path: path to a module (relative path)
            url: url of the module to track
        '''
        # correct path
        if not path:
            mod_name = url.split('/')[-1]
            # just in case there is an extension
            if '.' in mod_name:
                mod_name = mod_name.split('.')[0]
            path = os.path.join(relative_path(root), mod_name)
        super(ZoomModule, self).__init__(root, path, url)

    def sync(self, exact=False):
        '''
        Sync this module to the correct state

        args:
            exact: if True, ignore any tracking and sync to an exact state
        '''
        raise NotImplementedError()

    def track_branch(self, branch):
        '''
        Set this module to track the specified branch

        args:
            branch: branch to track
        '''
        raise NotImplementedError()

    def lock_commit(self, commit=None):
        '''
        Lock this module to the specified commit. If commit is None,
        lock to the last synced commit.

        args:
            commit: commit to lock to
        raises:
            ZoomError: on fail to set commit
        '''
        raise NotImplementedError()

    def lock_tag(self, tag):
        '''
        Lock this module to the specified tag

        args:
            tag: tag to lock to
        '''
        raise NotImplementedError()

    def execute(self, command):
        '''
        Execute a command inside of this module's context.

        args:
            command: command to execute (str)
        returns:
            retcode: return code of command
            stdout: stdout from command
            stderr: stderr from command
        '''
        self._log.info('running "{}" in {}'.format(command, str(self)))
        return run_command(command, self.abspath)



import os

from zoom_error import ZoomError
from zoom_resource import ZoomModule

GIT_DIR = '.git'

GIT_INIT = 'git init'
GIT_STATUS = 'git status'
GIT_ADD_REMOTE = 'git remote add origin {}'
GIT_FETCH = 'git fetch origin'
GIT_CHECKOUT = 'git checkout {}'
GIT_PULL = 'git pull origin {}'
GIT_REV_PARSE = 'git rev-parse HEAD'

class ZoomGitModule(ZoomModule):
    def __init__(self, root, path, url,
                 branch=None, tag=None, commit=None):
        '''
        Initialize zoom git module.

        args:
            root: root of the ZoomProject
            path: path to this module
            url: url for this dependency
            branch: branch to track
            tag: tag to lock to
            commit: commit to lock to
        '''
        super(ZoomGitModule, self).__init__(root, path, url)
        self.identifier = self.path
        self.rtype = 'git'
        self._set_revision(branch=branch, tag=tag, commit=commit)

    def _set_revision(self, branch=None, tag=None, commit=None):
        '''
        Single function to manage the logic of what pieces of information
        can be stored simultaneously.

        valid combos:
            1. branch
            2. tag
            3. commit
            4. branch + commit
            5. tag + commit
        the only invalid combos involve branch + tag, meaning simply that case
        must be disallowed. if that case is encountered, track branch.

        args:
            branch: branch to track
            tag: tag to lock to
            commit: commit to lock to
        '''
        # disallowed case, favor branch
        if branch and tag:
            tag = None
        self.branch = branch
        self.tag = tag
        self.commit = commit
        if not (self.branch or self.tag or self.commit):
            self.branch = 'master'

    def _is_git_repo(self):
        '''
        Check if this is a git repo

        returns:
            bool: true if git repo else false
        '''
        git_dir = os.path.join(self.abspath, GIT_DIR)
        return os.path.exists(git_dir)

    def _get_sha(self):
        '''
        Return the sha of the current git repo.

        returns:
            str: sha (git hash)
        raises:
            ZoomError: if this is not a git repo
        '''
        if not self._is_git_repo():
            raise ZoomError("not a git repo")
        _, stdout, _ = self.execute(GIT_REV_PARSE)
        return stdout.strip()

    def is_synced(self):
        '''
        Check if this module is already synced

        returns:
            bool: true if synced else false
        '''
        # Checking if this is a git repo is good enough
        return self._is_git_repo()

    def status(self):
        '''
        String representation of this module's status

        returns:
            str: status (stdout of git status)
        '''
        if self.is_synced():
            _, stdout, _ = self.execute(GIT_STATUS)
            return stdout
        else:
            return "repo not synced"

    def sync(self, exact=False):
        '''
        Sync this module to the correct state and store any updated info.

        args:
            exact: sync to the exact commit, even if tracking a branch
        raises:
            ZoomError: passes git stderr up if git commands fail
            ZoomError: if exact sync is called but there is no commit
        '''
        if exact and not self.commit:
            raise ZoomError('no exact commit for exact sync')

        # init repo if not already there
        if not self._is_git_repo():
            self._log.warning('initializing {}'.format(str(self)))
            self.execute(GIT_INIT)
            self.execute(GIT_ADD_REMOTE.format(self.url))

        # fetch changes
        retcode, _, stderr = self.execute(GIT_FETCH)
        if retcode != 0:
            raise ZoomError(stderr)

        # detemine whether this is tracking a branch or locking to a commit
        track = self.branch and not exact

        if track:
            # check out this branch and pull in changes
            retcode, _, stderr = self.execute(GIT_CHECKOUT.format(self.branch))
            if retcode != 0:
                raise ZoomError(stderr)
            retcode, _, stderr = self.execute(GIT_PULL.format(self.branch))
            if retcode != 0:
                raise ZoomError(stderr)
            self._log.warning('{} synced to {}'.format(str(self), self.branch))
        else:
            # check out the target commit or tag
            target_commit = self.commit or self.tag
            if exact:
                target_commit = self.commit
            retcode, _, stderr = self.execute(GIT_CHECKOUT.format(target_commit))
            if retcode != 0:
                raise ZoomError(stderr)
            self._log.warning('{} synced to {}'.format(str(self), target_commit))

        # update the commit in case it changed
        self.commit = self._get_sha()

    def track_branch(self, branch):
        '''
        Set this module to track the specified branch

        args:
            branch: branch to track
        '''
        self._set_revision(branch=branch)

    def lock_commit(self, commit=None):
        '''
        Lock this module to the specified commit. If commit is None,
        lock to the last synced commit.

        args:
            commit: commit to lock to
        raise:
            ZoomError: if no commit is available to lock to
        '''
        if not commit:
            commit = self.commit
        if not commit:
            raise ZoomError('no commit specified')
        self._set_revision(commit=commit)

    def lock_tag(self, tag):
        '''
        Lock this module to the specified tag

        args:
            tag: tag to lock to
        '''
        self._set_revision(tag=tag)

    def serialize(self):
        '''
        Serialize state of this module.

        returns:
            dict: dict representation of this module
        '''
        mdict = {
            'rtype': self.rtype,
            'path': self.path,
            'url': self.url
        }
        if self.branch:
            mdict['branch'] = self.branch
        if self.tag:
            mdict['tag'] = self.tag
        if self.commit:
            mdict['commit'] = self.commit

        return mdict



import os

from zoom_error import ZoomError
from zoom_resource import ZoomDependency
from zoom_io import download_file

class ZoomFileDep(ZoomDependency):
    def __init__(self, root, path, url,
                 filename=None):
        '''
        Initialize this file dependency.

        args:
            root: root of the ZoomProject
            path: path to this dependency
            url: url for this dependency
            filename: name to download file as
        '''
        super(ZoomFileDep, self).__init__(root, path, url)
        self.rtype = 'file'
        self.filename = filename
        if not filename:
            self.filename = self.url.split('/')[-1]
        self.identifier = os.path.join(self.path, self.filename)

    def is_synced(self):
        '''
        Check if this module is already synced

        returns:
            bool: true if synced else false
        '''
        return os.path.isfile(self.identifier)

    def status(self):
        '''
        String representation of this module's status

        returns:
            str: file downloaded or not downloaded
        '''
        if self.is_synced():
            return "file synced to: {}".format(self.identifier)
        else:
            return "file not synced"

    def sync(self):
        '''
        Sync this module to the correct state.

        raises:
            ZoomError: failed to download file
        '''
        try:
            download_file(self.url, self.abspath, self.filename)
        except IOError:
            raise ZoomError('failed to download file')

    def delete(self):
        '''
        Remove the downloaded file from the filesystem
        '''
        os.remove(self.identifier)

    def serialize(self):
        '''
        Serialize state of this dependency

        returns:
            dict: dict representation of this dependency
        '''
        return {
            'rtype': self.rtype,
            'path': self.path,
            'url': self.url,
            'filename': self.filename
        }



import os
import logging

import zoom_io
from zoom_error import ZoomError

from zoom_resource import ZoomDependency, ZoomModule
from zoom_git_module import ZoomGitModule
from zoom_file_dep import ZoomFileDep

resource_types = {
    'git': ZoomGitModule,
    'file': ZoomFileDep
}

class ZoomProject(object):
    '''
    Representation of an entire zoom project
    '''

    def __init__(self, create=False):
        '''
        Initialize zoom project using the cwd of the caller.

        args:
            create: will initialize a new project if true else just load
        raises:
            ZoomError: project already exists (if creating)
            ZoomError: no project exists (if loading)
            ZoomError: failed to write zoomfile
        '''
        self.name = None
        self.modules = []
        self.deps = []
        self.root = None
        self.path = os.getcwd()
        self.rel_path = None
        self._log = logging.getLogger('zoom')

        if create:
            self.name = os.path.basename(self.path)
            self.root = self.path
            try:
                zoom_io.create_proj_file(self.path, self.serialize())
            except IOError:
                raise ZoomError('zoom project already exists')
            self._save()
        else:
            state = None
            try:
                self.root, state = zoom_io.load_proj_file(self.path)
                self.rel_path = zoom_io.relative_path(self.root)
            except IOError:
                raise ZoomError('not inside a zoom project')
            self.deserialize(state)

    def _save(self):
        '''
        Save the state of this zoom project to the proj file

        raises:
            ZoomError: failed to save to disk
        '''
        try:
            zoom_io.save_proj_file(self.root, self.serialize())
        except IOError:
            raise ZoomError('failed to save zoom project')

    def sync(self, force=False, exact=False, limit=None,
             sync_cb=None, error_cb=None):
        '''
        Sync all resource associated with this zoom project to their
        state, as described in the zoomfile.

        args:
            force: sync resources from scratch (delete first)
            exact: sync modules to the exact commit (even if
                   they're tracking a branch)
            limit: list of relative paths to specifically sync
            sync_cb: callback on module sync (arg: ZoomModule)
            error_cb: callback on module sync error (arg: ZoomModule, error)
        '''
        # convert relative paths to abs paths
        if limit:
            limit = [os.path.abspath(l) for l in limit]

        # sync deps
        for dep in self.deps:
            if limit:
                if not in_limit(dep.abspath, limit):
                    self._log.warning('skipping {}...'.format(str(dep)))
                    continue
            try:
                if force:
                    self._log.warning('removing {}...'.format(str(dep)))
                    dep.delete()
                if callable(sync_cb): sync_cb(dep)
                dep.sync()
            except ZoomError as e:
                if callable(error_cb): error_cb(dep, e)

        # sync scm
        for module in self.modules:
            if limit:
                if not in_limit(module.abspath, limit):
                    self._log.warning('skipping {}...'.format(str(module)))
                    continue
            try:
                # pre-sync
                if force:
                    self._log.warning('removing {}...'.format(str(module)))
                    module.delete()
                if callable(sync_cb): sync_cb(module)
                module.sync(exact=exact)
            except ZoomError as e:
                if callable(error_cb): error_cb(module, e)

        # save any sync updates
        self._save()

    def get_resources(self):
        '''
        Get a list of all resources associated with this zoom project.

        returns:
            list: list of dependencies and modules
        '''
        return self.get_dependencies() + self.get_modules()

    def get_dependencies(self):
        '''
        Get a list of deps associated with this zoom project.

        returns:
            list: list of dependencies
        '''
        return self.deps

    def _get_module(self, path):
        '''
        Get the module that lives at path

        args:
            path: abs path to a module
        returns:
            ZoomModule: module with abspath, path. None if no match
        '''
        for m in self.modules:
            if m.abspath == path:
                return m
        return None

    def get_modules(self):
        '''
        Get a list of modules associated with this zoom project.

        returns:
            list: list of modules
        '''
        return self.modules

    def foreach_module(self, command, limit=None):
        '''
        Execute command in each zoom module

        args:
            command: command to execute (string)
            limit: limit command to run in modules in these paths
        returns:
            dict: dict of module to retcode, stdout, stderr
        '''
        # convert relative paths to abs paths
        if limit:
            limit = [os.path.abspath(l) for l in limit]

        outputs = {}
        for module in self.modules:
            if limit:
                if not in_limit(module.abspath, limit):
                    self._log.warning('skipping {}...'.format(str(module)))
                    continue
            self._log.warning('running in {}'.format(str(module)))
            retcode, stdout, stderr = module.execute(command)
            results = {
                'retcode': retcode,
                'stdout': stdout,
                'stderr': stderr
            }
            outputs[str(module)] = results
        return outputs

    def add_resource(self, rtype, args):
        '''
        Add a resource to this zoom project

        args:
            rtype: type of resource being added
        returns:
            ZoomResource: the new zoom resource
        raises:
            ZoomError: module name already in use
            ZoomError: failed to save zoomfile
        '''
        new_resource = load_resource(self.root, rtype, args)

        if self.root not in new_resource.abspath:
            raise ZoomError('\'{}\' is outside of zoom project'.format(new_resource.path))

        # validate no duplicate resource exists
        for resource in self.deps + self.modules:
            if new_resource.identifier.lower() == resource.identifier.lower():
                raise ZoomError('duplicate resource exists')

        # save to the correct place
        if isinstance(new_resource, ZoomDependency):
            self.deps.append(new_resource)
        else:
            self.modules.append(new_resource)

        # save the new resource
        self._save()
        return new_resource

    def track_branch(self, branch, limit):
        pass

    def lock_commit(self, path, commit=None):
        '''
        Lock the specified module to the specified commit. If commit is
        None, lock to the last synced commit.

        args:
            commit: commit to lock to
            path: relative path to a module
        return:
            ZoomModule: newly commit-locked module
        raises:
            ZoomError: if path does not point to a module
        '''
        abspath = os.path.abspath(path)
        module = self._get_module(abspath)
        if not module:
            raise ZoomError('{} is not a module'.format(path))

        module.lock_commit(commit)
        self._save()
        return module

    def lock_tag(self, tag, limit=None):
        '''
        Lock modules to a specified tag.

        args:
            tag: tag to lock to
            limit: list of paths to tag lock
        return:
            list: modules that got tag locked
        '''
        # convert relative paths to abs paths
        if limit:
            limit = [os.path.abspath(l) for l in limit]

        # list of modules that actually got locked
        locked = []

        # lock modules
        for module in self.modules:
            if limit:
                if not in_limit(module.abspath, limit):
                    continue
            module.lock_tag(tag)
            locked.append(module)

        self._save()
        return locked

    def serialize(self):
        '''
        Serialize state of this zoom project

        returns:
            dict: dict representation of this zoom project
        '''
        return {
            'name': self.name,
            'resources': {
                'dependencies': [d.serialize() for d in self.deps],
                'modules': [m.serialize() for m in self.modules]
            }
        }

    def deserialize(self, sdict):
        '''
        Load a zoom project from a serialized state

        args:
            sdict: dict representation of a zoom project
        raises:
            ZoomError: failed to deserialize
        '''
        try:
            self.name = sdict['name']
            self.deps = [load_resource(self.root,
                                       dep.pop('rtype'),
                                       dep)
                         for dep in sdict['resources']['dependencies']]
            self.modules = [load_resource(self.root,
                                          module.pop('rtype'),
                                          module)
                            for module in sdict['resources']['modules']]

        except ZoomError:
            raise ZoomError('failed to load zoomfile')


def in_limit(path, limit):
    '''
    Check if path is in limit.
    The criteria for this is a bit weird. A path is considered in limit if
    path has an _exact_ match in limit OR path's PARENT directory is a subdir
    of a path in limit.

    args:
        path: str path in question
        limit: list of paths to check against
    returns:
        bool: true if path is a subpath of any path in limit
    '''
    parent_dir = os.path.dirname(path)
    for l in limit:
        if path == l:
            return True
        if os.path.commonprefix([parent_dir, l]) == l:
            return True
    return False

def load_resource(root, rtype, rdict):
    '''
    Create the correct type of zoom resource from a serialized state

    args:
        root: path to the root of the zoom project
        rdict: type of ZoomResource to deserialize
        mdict: serialized representation of rdict ZoomResource
    returns:
        ZoomResource: rtype implementation of a ZoomResource
    raises:
        ZoomError: invalid rtype
    '''
    resource = None

    # lookup resource type
    try:
        resource = resource_types[rtype]
    except KeyError:
        raise ZoomError("invalid dependency type: {}". format(rtype))

    # get required args from rdict
    path = None
    url = None
    try:
        path = rdict.pop('path')
        url = rdict.pop('url')
    except KeyError as e:
        raise ZoomError('failed to parse resource')

    return resource(root, path, url, **rdict)



import os
import sys
import click
import logging

from zoom_project import ZoomProject
from zoom_error import ZoomError

def _stringify(o):
    '''
    Convert object o to str if unicode

    args:
        o: object to convert (dict, tuple, list, unicode)
    return:
        object matching the input type without unicode
    '''
    if type(o) is unicode:
        return str(o)
    elif type(o) is list or type(o) is tuple:
        return [_stringify(value) for value in o]
    elif type(o) is dict:
        return {key: _stringify(value) for key, value in o.iteritems()}
    else:
        return o

def load_zoom_proj(*args, **kwargs):
    '''
    Single source for an instance of a ZoomProject.
    Entirely added for testing. Click makes mocking a pain in the ass.
    '''
    return ZoomProject(*args, **kwargs)

def config_logger(verbosity):
    '''
    Configure logger for zoom with the specified verbosity.

    args:
        verbosity: int between 0 and 3. larger number = more verbose
    '''
    lvl = (4 - verbosity) * 10
    # configure logger
    logger = logging.getLogger('zoom')
    logger.setLevel(lvl)

    ch = logging.StreamHandler()
    ch.setLevel(lvl)
    logger.addHandler(ch)

#
# CLI entrypoint
#

@click.group()
@click.option('-v','--verbose', count=True)
@click.pass_context
def cli(ctx, verbose):
    # set up logger
    config_logger(verbose)

    # initialize zoom_proj based on command
    zoom_proj = None
    if ctx.invoked_subcommand != 'init':
        try:
            zoom_proj = load_zoom_proj()
        except ZoomError as e:
            print 'fatal:', e
            sys.exit(1)
    ctx.obj = zoom_proj

#
# INIT
#

@cli.command(help='create a new, empty zoom project. '
             'generates a zoomfile (zoom.yml)')
def init():
    try:
        load_zoom_proj(create=True)
        print "zoom project created"
    except ZoomError as e:
        print 'fatal:', e
        sys.exit(1)

#
# Include commands
#

def _add_resource(zoom_proj, rtype, args):
    '''
    Wrapper for zoom_proj.add_resource to ensure args are in
    the correct format.
    '''
    args = _stringify(args)
    try:
        resource = zoom_proj.add_resource(rtype, args)
        print "added resource {}".format(str(resource))
    except ZoomError as e:
        print 'fatal:', e
        sys.exit(1)

@cli.group(help="include a resource in this zoom project")
@click.pass_obj
def include(zoom_proj):
    pass

@include.command('git', help='include a git repo.')
@click.option('-c', '--commit', metavar='<commit>',
              help='sha to lock to')
@click.option('-b', '--branch', metavar='<branch>',
              help='branch to track')
@click.option('-t', '--tag', metavar='<tag>',
              help='tag to lock to')
@click.argument('url')
@click.argument('path', required=False)
@click.pass_obj
def add_git(zoom_proj, commit, branch, tag, url, path):
    args = {'commit': commit,
            'branch': branch,
            'tag': tag,
            'url': url,
            'path': path}
    _add_resource(zoom_proj, 'git', args)

@include.command('file', help='include a file.')
@click.option('-f', '--filename', metavar='<filename>',
              help='filename to resolve to')
@click.argument('url')
@click.argument('path', required=False)
@click.pass_obj
def add_file(zoom_proj, filename, url, path):
    args = {'filename': filename,
            'url': url,
            'path': path}
    _add_resource(zoom_proj, 'file', args)

#
# Foreach
#

@cli.command(help='run a command across multiple modules.')
@click.argument('command')
@click.argument('limit', nargs=-1)
@click.pass_obj
def foreach(zoom_proj, command, limit):
    command = _stringify(command)
    limit = _stringify(limit)
    results = zoom_proj.foreach_module(command, limit=limit)
    for module, result in results.iteritems():
        print str(module)
        err = result['retcode'] != 0
        text = result['stderr'] if err else result['stdout']
        print '\t' + ('\n\t'.join(text.split('\n')))

#
# LIST
#

@cli.command(help='list all resources in this zoom project.')
@click.pass_obj
def list(zoom_proj):
    resources = zoom_proj.get_resources()
    for resource in resources:
        print str(resource)

#
# STATUS
#

@cli.command(help='display the current status for all resources.')
@click.pass_obj
def status(zoom_proj):
    resources = zoom_proj.get_resources()
    for resource in resources:
        print str(resource)
        print '\t' + ('\n\t'.join(resource.status().split('\n')))

#
# SYNC
#

def sync_cb(r):
    '''
    Callback - print start status
    '''
    print 'syncing {}...'.format(str(r))


def sync_err_cb(r, e):
    '''
    Callback - print sync error
    '''
    print 'error syncing {} - {}'.format(str(r), str(e))

@cli.command(help='download resources to thier specified state and location.')
@click.option('-e', '--exact', is_flag=True,
              help='sync modules to the exact commit set in the zoomfile. '
              'this will override branch tracking.')
@click.option('-f', '--force', is_flag=True,
              help='delete and re-sync resources')
@click.argument('limit', nargs=-1)
@click.pass_obj
def sync(zoom_proj, exact, force, limit):
    limit = _stringify(limit)
    zoom_proj.sync(force=force,
                   exact=exact,
                   limit=limit,
                   sync_cb=sync_cb,
                   error_cb=sync_err_cb)
    print 'sync complete'

#
# REV-LOCK
#

@cli.command('rev-lock', help='lock a module to a commit. If no commit '
             'specified, use the last synced commit')
@click.argument('module')
@click.argument('commit', required=False)
@click.pass_obj
def rev_lock(zoom_proj, module, commit):
    module = _stringify(module)
    commit = _stringify(commit)
    try:
        module = zoom_proj.lock_commit(module, commit)
        print 'locked {}@{}'.format(str(module), module.commit)
    except ZoomError as e:
        print 'fatal:', e
        sys.exit(1)

#
# TAG-LOCK
#

@cli.command('tag-lock', help='lock modules to a specified tag.')
@click.argument('tag')
@click.argument('limit', nargs=-1)
@click.pass_obj
def tag_lock(zoom_proj, tag, limit):
    tag = _stringify(tag)
    limit = _stringify(limit)
    try:
        modules = zoom_proj.lock_tag(tag, limit=limit)
        for module in modules:
            print 'locked {}@{}'.format(str(module), module.tag)
    except ZoomError as e:
        print 'fatal:', e
        sys.exit(1)



import os
import yaml
import urllib
import contextlib
import subprocess
import logging

from subprocess import PIPE

from zoom_error import ZoomError

ZOOM_FILE = "zoom.yml"

_log = logging.getLogger('zoom')
# os-independent root directory
root_dir = os.path.abspath(os.sep)

def relative_path(base):
    '''
    Get path of cwd relative to base path.

    args:
        base: path to get path relative to
    '''
    return os.getcwd().replace(base, '').strip(os.sep)

def find_proj_file(path):
    '''
    Search path and parents directories for a project file
    until one is found. Or not.

    args:
        path: path to start looking for a project file in
    returns:
        string: full path of proj file
    raises:
        IOError: file could not be found in path or parent dirs
    '''
    projfile = None
    search_path = path
    while True:
        projfile = os.path.join(search_path, ZOOM_FILE)

        if os.path.exists(projfile):
            return projfile

        search_path = os.path.dirname(search_path)
        if search_path == root_dir:
            raise IOError()

def in_zoom_project(path):
    '''
    Checks if path is within a zoom project.

    args:
        path: path in question
    returns:
        bool: true if in zoom project else false
    '''
    try:
        find_proj_file(path)
    except IOError:
        return False
    return True

def in_zoom_project_root(path):
    '''
    Checks if path is the root of a zoom project

    args:
        path: path in question
    returns:
        bool: true if in root of zoom project else false
    '''
    try:
        projfile = find_proj_file(path)
        return os.path.dirname(projfile) == path
    except IOError:
        return False

def create_proj_file(root, content):
    '''
    Create new project file, if it does not already exist.

    args:
        root: path to create project file in
        content: dictionary representing the contents of the file
    raises:
        IOError: project file already exists
    '''
    if in_zoom_project_root(root):
        raise IOError()
    else:
        save_proj_file(root, content)

def save_proj_file(root, content):
    '''
    Write content to a zoom project file.

    args:
        root: root of a zoom project
        content: dictionary representing the contents of the file
    raises:
        IOError: failed to write to file
    '''
    projfile = os.path.join(root, ZOOM_FILE)
    with open(projfile, 'w+') as f:
        f.write(yaml.dump(content, default_flow_style=False))

def load_proj_file(path):
    '''
    Loads contents of a zoom project file.

    args:
        path: path within a zoom project
    returns:
        string: dir where project file was found
        dict: contents of project file
    raises:
        IOError: file could not be found
    '''
    projfile = None
    try:
        projfile = find_proj_file(path)
    except IOError:
        raise IOError()
    with open(projfile, 'r') as f:
        return (os.path.dirname(projfile), yaml.load(f.read()))

@contextlib.contextmanager
def path_context(path):
    '''
    Context manager for a given path. cd in and cd out.
    Creates directory if it does not exist.

    args:
        path: path to cd into
    '''
    prev_cwd = os.getcwd()
    if not os.path.exists(path):
        os.makedirs(path)
    os.chdir(path)
    try:
        yield
    finally:
        os.chdir(prev_cwd)

def download_file(url, path, filename):
    '''
    Download file to path with name filename if it DNE.
    Will create directories.

    args:
        url: url to download
        path: dir to place file
        filename: name of file
    raises:
        IOError: failed to download file
    '''
    if not os.path.exists(path):
        os.makedirs(path)
    filepath = os.path.join(path, filename)
    if not os.path.exists(filepath):
        # raises IOError on fail
        urllib.urlretrieve(url, filepath)

def run_command(command, path):
    '''
    Run command in path.

    args:
        command: command to run in list form
        path: dir to run in
    returns:
        int: return code of command
        str: stdout from command
        str: stderr from command
    raises:
        None
    '''
    # this could use Popen's cwd arg, but path_context will
    # handle the case the directory doesn't already exist
    with path_context(path):
        _log.debug('running {} in {}'.format(''.join(command), path))
        process = subprocess.Popen(command,
                                   shell=True,
                                   stdout=PIPE,
                                   stderr=PIPE)
        out, err = process.communicate()
        retcode = process.returncode
        if retcode == 0:
            _log.debug(out.strip())
        else:
            _log.debug(err.strip())
        return (retcode, out.strip(), err.strip())

def is_windows():
    '''
    Check if running on Windows.

    returns:
        bool: true if executed on windows, else false
    '''
    return os.name == 'nt'

def posixify(string):
    '''
    Converts Windows-badness to posixy-goodness.
    Basically just swaps '\' for '/'

    args:
        string: str to convert
    returns:
        str: converted string
    '''
    if is_windows():
        return string.replace('\\', '/')
    else:
        return string



from setuptools import setup
from zoom.__version__ import __version__

setup(
    name='zoom',
    packages=['zoom'],
    version=__version__,
    description='Distributed project management',
    author='Sunny Sagar',
    author_email='trexinabluecape@gmail.com',
    url='https://github.com/ssagar0/zoom',
    install_requires=['click', 'pyaml'],
    entry_points={
        'console_scripts': [
            'zoom=zoom:entrypoint',
        ],
    },
    license='MIT',
    keywords='git repo',
    classifiers=[
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 2.7',
        'Topic :: Software Development :: Version Control'
    ],
)
